import time
import boto.ec2
import boto.ec2.autoscale
from boto.ec2.autoscale import LaunchConfiguration
from boto.ec2.autoscale import AutoScalingGroup
from boto.ec2.autoscale import ScalingPolicy
import boto.ec2.cloudwatch
from . import iquery, metrics

class AutoScaler:	
	def __init__ (self, name, region = 'us-east-1'):
		self.name = name
		self.region = region
		self.launch_config = None
		self.as_group = None
		self.conn = boto.ec2.autoscale.connect_to_region(self.region)
		
	def create_launch_config (self, param):		
		if len (self.conn.get_all_launch_configurations (names = [self.name])):
			print ('launch configuration already exists')
			return	
		param ["name"] = self.name
		lc = LaunchConfiguration(**param)
		self.conn.create_launch_configuration (lc)
		self.launch_config = lc
		
	def create_auto_scaling_group (self, min_size = 1, max_size = 2):		
		if len (self.conn.get_all_groups (names = [self.name])):
			print ('auto-scaling group already exists')
			return			
		
		if self.launch_config is None:
			try:
				self.launch_config = self.conn.get_all_launch_configurations (names = [self.name])[0]
			except IndexError:
				print ('launch configuration not exists')
				return
			
		self.as_group = AutoScalingGroup(
			group_name=self.name,
			load_balancers=['skitai-app'],
			availability_zones=[self.region + 'a', self.region + 'b', self.region + 'c', self.region + 'd', self.region + 'e', self.region + 'f'],
			launch_config = self.launch_config, 
			min_size = min_size, 
			max_size = max_size,
			connection = self.conn
		)
		self.conn.create_auto_scaling_group (self.as_group)		
	
	def make_policies (self):
		scale_up_policy = ScalingPolicy(
			name='scale-up', adjustment_type='ChangeInCapacity',
			as_name=self.name, scaling_adjustment=1, cooldown=180
		)
		scale_down_policy = ScalingPolicy(
			name='scale-down', adjustment_type='ChangeInCapacity',
			as_name=self.name, scaling_adjustment=-1, cooldown=180
		)
		self.conn.create_scaling_policy(scale_up_policy)
		self.conn.create_scaling_policy(scale_down_policy)

	def get_policies (self):
		scale_up_policy = self.conn.get_all_policies (
			as_group=self.name, policy_names=['scale-up'])[0]            
		scale_down_policy = self.conn.get_all_policies (
			as_group=self.name, policy_names=['scale-down'])[0]
		return scale_up_policy, scale_down_policy

	def create_cpu_alram (self, low = 40, high = 70):
		cloudwatch = boto.ec2.cloudwatch.connect_to_region(self.region)
		alarm_dimensions = {"AutoScalingGroupName": self.name}	
		scale_up_policy, scale_down_policy = self.get_policies ()
		
		alarm = metrics.create_alram (
			'%s-scale-up-on-cpu' % self.name,
			'CPUUtilization',	'>', str (high),
			alarm_dimensions,
			scale_up_policy.policy_arn
		)
		cloudwatch.create_alarm (alarm)
		
		alarm = metrics.create_alram (
			'%s-scale-down-on-cpu' % self.name,
			'CPUUtilization',	'<', str (low),		
			alarm_dimensions,
			scale_down_policy.policy_arn		
		)
		cloudwatch.create_alarm (alarm)
	
	def create_credit_alram (self, low = 20, high = 60):
		cloudwatch = boto.ec2.cloudwatch.connect_to_region(self.region)
		alarm_dimensions = {"AutoScalingGroupName": self.name}	
		scale_up_policy, scale_down_policy = self.get_policies ()	
		
		alarm = metrics.create_alram (
			'%s-scale-up-on-credit' % self.name,
			'CPUCreditBalance',	'<', str (low),
			alarm_dimensions,
			scale_up_policy.policy_arn
		)
		cloudwatch.create_alarm (alarm)
		
		alarm = metrics.create_alram (
			'%s-scale-down-on-credit' % self.name,
			'CPUCreditBalance',	'>', str (high),
			alarm_dimensions,
			scale_down_policy.policy_arn
		)
		cloudwatch.create_alarm (alarm)
	
	def get_instances (self):
		return [inst for inst in iquery.get_instances_by_auto_scaling_group (self.name)]
	
	def get_instance_ids (self):
		return [h.id for h in self.get_instances ()]
			
	def is_empty (self):
		runnings = [1 for inst in iquery.get_instances_by_auto_scaling_group (self.name) if inst.update () != 'terminated']
		return not len (runnings)
		
	def remove (self):		
		ags = self.conn.get_all_groups (names = [self.name])
		if not ags:
			print ('auto scaling group not exists')
			return		
		ag = ags [0]
		ag.shutdown_instances ()
		
		while not self.is_empty ():
			print ('wait for all instances termination...', runnings)
			time.sleep (10)			
		ag.delete()
		
		lcs = self.conn.get_all_launch_configurations (names = [self.name])
		if not lcs:
			print ('launch configuration not exists')
			return		
		lc = lcs [0]
		lc.delete()	



# -*- Mode: Python; tab-width: 4 -*-

RCS_ID = '$Id: producers.py,v 1.10 1999/02/01 03:08:46 rushing Exp $'

import string
import time
import io
import gzip
from . import compressors, strutil
import mimetypes
import os
from collections import Iterable, deque
import asynchat

"""
A collection of producers.
Each producer implements a particular feature:  They can be combined
in various ways to get interesting and useful behaviors.

For example, you can feed dynamically-produced output into the compressing
producer, then wrap this with the 'chunked' transfer-encoding producer.
"""

SIZE_BUFFER = asynchat.async_chat.ac_out_buffer_size

class simple_producer:
	"producer for a string"
	def __init__ (self, data, buffer_size = SIZE_BUFFER):
		self.data = data
		self.buffer_size = buffer_size
		self.proxsize = len (data)
	
	def get_size (self):
		return self.proxsize
		
	def more (self):
		if len (self.data) > self.buffer_size:
			result = self.data[:self.buffer_size]
			self.data = self.data[self.buffer_size:]
			return result
		else:
			result = self.data
			self.data = b''
			return result

class list_producer (simple_producer):
	def __init__ (self, data):
		self.data = data
		self.proxsize = sum ([len (each) for each in self.data])
		
	def more (self):
		if not self.data:
			return b''			
		data = self.data.pop (0)
		if strutil.is_encodable (data):
			return data.encode ("utf8")
		return data

class iter_producer (list_producer):
	def __init__ (self, data):
		self.data = data
		self.proxsize = -1
	
	def more (self):
		try:
			data = next (self.data)
			if strutil.is_encodable (data):
				return data.encode ("utf8")
			return data
		except StopIteration:		
			return b""
	
class closing_stream_producer (simple_producer):
	def __init__ (self, data, buffer_size = SIZE_BUFFER):
		self.data = data
		self.buffer_size = buffer_size
		self.closed = False
		self.proxsize = hasattr (data, "get_size") and data.get_size () or -1
			
	def more (self):
		if self.closed: 
			return b""
		data = self.data.read (self.buffer_size)		
		if not data:
			self.close ()
		return data
	
	def close (self):
		if self.closed: return			
		try: self.data.close ()
		except AttributeError: pass	
		self.closed = True
	
class scanning_producer:
	"like simple_producer, but more efficient for large strings"
	def __init__ (self, data, buffer_size = SIZE_BUFFER):
		self.data = data
		self.buffer_size = buffer_size
		self.proxsize = len (self.data)
		self.pos = 0

	def get_size (self):
		return self.proxsize
		
	def more (self):
		if self.pos < len(self.data):
			lp = self.pos
			rp = min (
				len(self.data),
				self.pos + self.buffer_size
				)
			result = self.data[lp:rp]
			self.pos = self.pos + len(result)
			return result
		else:
			return b''

class lines_producer:
	"producer for a list of lines"

	def __init__ (self, lines):
		self.lines = lines
	
	def get_size (self):
		return sum ([len (each) for each in self.data])
		
	def ready (self):
		return len(self.lines)

	def more (self):
		if self.lines:
			chunk = self.lines[:50]
			self.lines = self.lines[50:]
			return b'\r\n'.join (chunk) + b'\r\n'
		else:
			return b''

class buffer_list_producer:
	"producer for a list of buffers"

	# i.e., data == string.join (buffers, '')
	
	def __init__ (self, buffers):
		self.index = 0
		self.buffers = buffers
		
	def get_size (self):
		return len (self.buffers)
		
	def more (self):
		if self.index >= len(self.buffers):
			return b''
		else:
			data = self.buffers[self.index]
			self.index = self.index + 1
			return data

class file_producer:
	"producer wrapper for file[-like] objects"

	# match http_channel's outgoing buffer size
	def __init__ (self, file, buffer_size = SIZE_BUFFER, proxsize = -1):
		self.done = 0
		self.file = file
		self.buffer_size = buffer_size
		self.proxsize = proxsize
	
	def get_size (self):
		if self.proxsize != -1:
			return self.proxsize
		# possible io object
		try:
			return os.path.getsize (self.file.name)
		except:
			return -1	
			
	def more (self):
		if self.done:
			return b''
		else:
			data = self.file.read (self.buffer_size)			
			if not data:
				self.file.close()
				del self.file
				self.done = 1
				return b''
			else:
				return data

# A simple output producer.  This one does not [yet] have
# the safety feature builtin to the monitor channel:  runaway
# output will not be caught.

# don't try to print from within any of the methods
# of this object.

class output_producer:
	"Acts like an output file; suitable for capturing sys.stdout"
	
	def __init__ (self):
		self.data = ''
		self.proxsize = 0
	
	def get_size (self):
		return self.proxsize
				
	def write (self, data):
		lines = data.split (b'\n')
		data = b'\r\n'.join (lines)
		self.proxsize += len (data)
		self.data = self.data + data
		
	def writeline (self, line):
		self.proxsize += (len (line) + 2)
		self.data = self.data + line + b'\r\n'
		
	def writelines (self, lines):
		d = b'\r\n'.join (lines)
		self.proxsize += (len (d) + 2)
		self.data = self.data + d + b'\r\n'

	def ready (self):
		return (len (self.data) > 0)

	def flush (self):
		pass

	def softspace (self, *args):
		pass

	def more (self):
		if self.data:
			result = self.data[:512]
			self.data = self.data[512:]
			return result
		else:
			return b''

		
class composite_producer:
	"combine a fifo of producers into one"
	def __init__ (self, producers):
		self.producers = producers		
		self.override ()
	
	def get_size (self):
		return self.estimate_size ()
		
	def estimate_size (self):
		size = 0
		for p in self.producers:
			s = p.get_size () 
			if s == -1:
				return -1
			size += s
		return size
			
	def override (self):
		if len (self.producers) == 0:
			return
		p = self.producers.first ()
		if hasattr (p, "ready"):
			self.ready = p.ready
			p.ready = None
		elif hasattr (self, 'ready'):
			del self.ready
						
	def more (self):
		while len (self.producers):
			p = self.producers.first()
			d = p.more()			
			if d:
				return d
			else:
				self.producers.pop()
				self.override ()
		else:
			return b''

class globbing_producer:
	"""
	'glob' the output from a producer into a particular buffer size.
	helps reduce the number of calls to send().  [this appears to
	gain about 30% performance on requests to a single channel]
	"""

	def __init__ (self, producer, buffer_size = SIZE_BUFFER):
		self.producer = producer
		if hasattr (self.producer, "ready"):
			raise TypeError ("globbing producer cannot have ready methods")
		self.buffer = b''
		self.buffer_size = buffer_size
	
	def get_size (self):
		return self.producer.get_size ()
		
	def more (self):
		while len(self.buffer) < self.buffer_size:
			data = self.producer.more()
			if data:
				self.buffer = self.buffer + data
			else:
				break
		r = self.buffer
		self.buffer = b''
		return r


class hooked_producer (globbing_producer):
	"""
	A producer that will call <function> when it empties,.
	with an argument of the number of bytes produced.  Useful
	for logging/instrumentation purposes.
	"""
	def __init__ (self, producer, function):
		self.producer = producer
		self.function = function
		self.bytes = 0		
		self.override ()
	
	def override (self):	
		if hasattr (self.producer, "ready"):
			self.ready = self.producer.ready
			self.producer.ready = None
				
	def more (self):
		if self.producer:
			result = self.producer.more()			
			if not result:				
				self.producer = None				
				self.function (self.bytes)
			else:
				self.bytes = self.bytes + len(result)
			return result
		else:
			return b''

# HTTP 1.1 emphasizes that an advertised Content-Length header MUST be
# correct.  In the face of Strange Files, it is conceivable that
# reading a 'file' may produce an amount of data not matching that
# reported by os.stat() [text/binary mode issues, perhaps the file is
# being appended to, etc..]  This makes the chunked encoding a True
# Blessing, and it really ought to be used even with normal files.
# How beautifully it blends with the concept of the producer.

class chunked_producer (hooked_producer):
	"""A producer that implements the 'chunked' transfer coding for HTTP/1.1.
	Here is a sample usage:
		request['Transfer-Encoding'] = 'chunked'
		request.push (
			producers.chunked_producer (your_producer)
			)
		request.done()
	"""

	def __init__ (self, producer, footers=None):
		self.producer = producer
		self.footers = footers
		self.override ()
			
	def more (self):
		if self.producer:
			data = self.producer.more()
			if data:
				dlen = '%x' % len (data)
				return dlen.encode ("utf8") + b"\r\n" + data + b'\r\n'
			else:				
				self.producer = None
				if self.footers:
					return b'\r\n'.join (
						[b'0'] + self.footers
					) + b'\r\n\r\n'
				else:
					return b'0\r\n\r\n'
		else:
			return b''

# Unfortunately this isn't very useful right now (Aug 97), because
# apparently the browsers don't do on-the-fly decompression.  Which
# is sad, because this could _really_ speed things up, especially for
# low-bandwidth clients (i.e., most everyone).

try:
	import zlib
except ImportError:
	zlib = None

class compressed_producer (hooked_producer):
	"""
	Compress another producer on-the-fly, using ZLIB
	[Unfortunately, none of the current browsers seem to support this]
	"""

	# Note: It's not very efficient to have the server repeatedly
	# compressing your outgoing files: compress them ahead of time, or
	# use a compress-once-and-store scheme.  However, if you have low
	# bandwidth and low traffic, this may make more sense than
	# maintaining your source files compressed.
	#
	# Can also be used for compressing dynamically-produced output.

	def __init__ (self, producer, level=6):
		self.producer = producer
		self.compressor = zlib.compressobj (level, zlib.DEFLATED)
		self.override ()
	
	def more (self):
		if self.producer:
			cdata = b''
			# feed until we get some output
			while not cdata:
				data = self.producer.more()
				if not data:
					self.producer = None
					return self.compressor.flush()
				else:
					cdata = self.compressor.compress (data)					
			return cdata
		else:
			return b''


class gzipped_producer (compressed_producer):
	def __init__ (self, producer, level=5):
		self.producer = producer
		self.compressor = compressors.GZipCompressor (level)		
		self.override ()

	
class escaping_producer:

	"A producer that escapes a sequence of characters"
	" Common usage: escaping the CRLF.CRLF sequence in SMTP, NNTP, etc..."

	def __init__ (self, producer, esc_from='\r\n.', esc_to='\r\n..'):
		self.producer = producer
		self.esc_from = esc_from
		self.esc_to = esc_to
		self.buffer = b''
		from asynchat import find_prefix_at_end
		self.find_prefix_at_end = find_prefix_at_end		
		self.override ()
	
	def get_size (self):
		return self.producer.get_size ()
		
	def more (self):
		esc_from = self.esc_from
		esc_to   = self.esc_to

		buffer = self.buffer + self.producer.more()

		if buffer:
			buffer = string.replace (buffer, esc_from, esc_to)
			i = self.find_prefix_at_end (buffer, esc_from)
			if i:
				# we found a prefix
				self.buffer = buffer[-i:]
				return buffer[:-i]
			else:
				# no prefix, return it all
				self.buffer = b''
				return buffer
		else:
			return buffer

class fifo:
	def __init__(self, list=None):
		if not list:
			self.list = deque()
		else:
			self.list = deque(list)
	
	def get_estimate_content_length (self):
		cl = 0
		for p in self.list:
			s = p.get_size ()
			if s == -1:
				return -1				
			cl += s	
		return cl
		
	def __len__(self):
		return len(self.list)

	def is_empty(self):
		return not self.list

	def first(self):
		return self.list[0]

	def push(self, data):
		self.list.append(data)
	
	def push_front (self, data):
		self.push(data)
		self.list.rotate (1)
		
	def pop(self):
		if self.list:
			return (1, self.list.popleft())
		else:
			return (0, None)
	
	def clear (self):
		self.list.clear ()
				


"""
 $Id: Lib.py,v 1.11 2002/03/19 13:05:02 anthonybaxter Exp $

 This file is part of the pydns project.
 Homepage: http://pydns.sourceforge.net

 This code is covered by the standard Python License.

 Library code. Largely this is packers and unpackers for various types.
"""

#
#
# See RFC 1035:
# ------------------------------------------------------------------------
# Network Working Group									 P. Mockapetris
# Request for Comments: 1035										   ISI
#															November 1987
# Obsoletes: RFCs 882, 883, 973
#
#			 DOMAIN NAMES - IMPLEMENTATION AND SPECIFICATION
# ------------------------------------------------------------------------


import types
import sys
from . import Type, Class, Opcode, Status
from .Base import DNSError

PY_MAJOR_VERSION = sys.version_info.major

class UnpackError(DNSError): pass
class PackError(DNSError): pass

# Low-level 16 and 32 bit integer packing and unpacking

from struct import pack as struct_pack
from struct import unpack as struct_unpack

def pack16bit(n):
	return struct_pack('!H', n)

def pack32bit(n):
	return struct_pack('!L', n)

def unpack16bit(s):
	return struct_unpack('!H', s)[0]

def unpack32bit(s):
	return struct_unpack('!L', s)[0]

def addr2bin(addr):
	if type(addr) == type(0): return addr
	bytes = addr.split('.')
	if len(bytes) != 4: raise ValueError('bad IP address')
	n = 0
	for byte in bytes: n = n<<8 | int(byte)
	return n

def bin2addr(n):
	return '%d.%d.%d.%d' % ((n>>24)&0xFF, (n>>16)&0xFF,
				  (n>>8)&0xFF, n&0xFF)


# Packing class

class Packer:
	" packer base class. supports basic byte/16bit/32bit/addr/string/name "
	def __init__(self):
		self.buf = b''
		self.index = {}
	def getbuf(self):
		return self.buf
	def addbyte(self, c):
		if len(c) != 1: raise TypeError('one character expected')
		self.buf = self.buf + c
	def addbytes(self, bytes):
		self.buf = self.buf + bytes
	def add16bit(self, n):
		self.buf = self.buf + pack16bit(n)
	def add32bit(self, n):
		self.buf = self.buf + pack32bit(n)
	def addaddr(self, addr):
		n = addr2bin(addr)
		self.buf = self.buf + pack32bit(n)
	def addstring(self, s):
		if len(s) > 255:
			raise ValueError("Can't encode string of length "+ \
							"%s (> 255)"%(len(s)))
		self.addbyte(chr(len(s)))
		self.addbytes(s)
	def addname(self, name):
		# Domain name packing (section 4.1.4)
		# Add a domain name to the buffer, possibly using pointers.
		# The case of the first occurrence of a name is preserved.
		# Redundant dots are ignored.
		list = []
		for label in name.split (b'.'):
			if label:
				if len(label) > 63:
					raise PackError('label too long')
				list.append(label)
		keys = []
		for i in range(len(list)):
			key = b".".join (list[i:]).upper()
			keys.append(key)
			if key in self.index:
				pointer = self.index[key]
				break
		else:
			i = len(list)
			pointer = None
		# Do it into temporaries first so exceptions don't
		# mess up self.index and self.buf
		buf = b''
		offset = len(self.buf)
		index = []
		for j in range(i):
			label = list[j]
			n = len(label)
			if offset + len(buf) < 0x3FFF:
				index.append((keys[j], offset + len(buf)))
			else:
				print('DNS.Lib.Packer.addname: %S warning: pointer too big' % name)
			buf = buf + (chr(n).encode ("utf8") + label)
		if pointer:
			buf = buf + pack16bit(pointer | 0xC000)
		else:
			buf = buf + b'\0'
		self.buf = self.buf + buf
		for key, value in index:
			self.index[key] = value
	def dump(self):
		keys = list(self.index.keys())
		keys.sort()
		print('-'*40)
		for key in keys:
			print('%20s %3d' % (key, self.index[key]))
		print('-'*40)
		space = 1
		for i in range(0, len(self.buf)+1, 2):
			if self.buf[i:i+2] == '**':
				if not space: print()
				space = 1
				continue
			space = 0
			for c in self.buf[i:i+2]:
				if 20 < c < 177:
					print('%4d  %c' % (i, c))
				else:
					print('%4d %2d' % (i, c))
			print("\n")
		print('-'*40)


# Unpacking class


class Unpacker:
	def __init__(self, buf):
		self.buf = buf		
		self.offset = 0
	def getbyte(self):
		if self.offset > len(self.buf):
			raise UnpackError("Ran off end of data")
		c = self.buf[self.offset]
		self.offset = self.offset + 1
		return c
	def getbytes(self, n):
		s = self.buf[self.offset : self.offset + n]
		if len(s) != n: raise UnpackError('not enough data left')
		self.offset = self.offset + n
		return s
	def get16bit(self):
		return unpack16bit(self.getbytes(2))
	def get32bit(self):
		return unpack32bit(self.getbytes(4))
	def getaddr(self):
		return bin2addr(self.get32bit())
	def getstring(self):
		return self.getbytes(ord(self.getbyte()))
	def getname(self):
		return self._getname ().decode ("utf8")
	def _getname(self):
		# Domain name unpacking (section 4.1.4)
		if PY_MAJOR_VERSION >= 3:
			i = self.getbyte()
		else:	
			c = self.getbyte()
			i = ord(c)		
		if i & 0xC0 == 0xC0:
			if PY_MAJOR_VERSION >= 3:
				j = self.getbyte()
			else:	
				d = self.getbyte()
				j = ord(d)
			pointer = ((i<<8) | j) & ~0xC000
			save_offset = self.offset
			try:
				self.offset = pointer
				domain = self._getname()
			finally:
				self.offset = save_offset
			return domain
		if i == 0:
			return b''
		domain = self.getbytes(i)
		remains = self._getname()
		if not remains:
			return domain
		else:
			return (domain + b'.' + remains)


# Test program for packin/unpacking (section 4.1.4)

def testpacker():
	N = 2500
	R = list(range(N))
	# See section 4.1.4 of RFC 1035
	for i in R:
		p = Packer()
		p.addaddr('192.168.0.1')
		p.addbytes('*' * 20)
		p.addname('f.ISI.ARPA')
		p.addbytes('*' * 8)
		p.addname('Foo.F.isi.arpa')
		p.addbytes('*' * 18)
		p.addname('arpa')
		p.addbytes('*' * 26)
		p.addname('')

	#p.dump()
	u = Unpacker(p.buf)
	u.getaddr()
	u.getbytes(20)
	u.getname()
	u.getbytes(8)
	u.getname()
	u.getbytes(18)
	u.getname()
	u.getbytes(26)
	u.getname()

	for i in R:
		u = Unpacker(p.buf)

		res = (u.getaddr(),
			   u.getbytes(20),
			   u.getname(),
			   u.getbytes(8),
			   u.getname(),
			   u.getbytes(18),
			   u.getname(),
			   u.getbytes(26),
			   u.getname())

	#for item in res: print item


# Pack/unpack RR toplevel format (section 3.2.1)

class RRpacker(Packer):
	def __init__(self):
		Packer.__init__(self)
		self.rdstart = None
	def addRRheader(self, name, type, klass, ttl, *rest):
		self.addname(name)
		self.add16bit(type)
		self.add16bit(klass)
		self.add32bit(ttl)
		if rest:
			if rest[1:]: raise TypeError('too many args')
			rdlength = rest[0]
		else:
			rdlength = 0
		self.add16bit(rdlength)
		self.rdstart = len(self.buf)
	def patchrdlength(self):
		rdlength = unpack16bit(self.buf[self.rdstart-2:self.rdstart])
		if rdlength == len(self.buf) - self.rdstart:
			return
		rdata = self.buf[self.rdstart:]
		save_buf = self.buf
		ok = 0
		try:
			self.buf = self.buf[:self.rdstart-2]
			self.add16bit(len(rdata))
			self.buf = self.buf + rdata
			ok = 1
		finally:
			if not ok: self.buf = save_buf
	def endRR(self):
		if self.rdstart is not None:
			self.patchrdlength()
		self.rdstart = None
	def getbuf(self):
		if self.rdstart is not None: self.patchrdlength()
		return Packer.getbuf(self)
	# Standard RRs (section 3.3)
	def addCNAME(self, name, klass, ttl, cname):
		self.addRRheader(name, Type.CNAME, klass, ttl)
		self.addname(cname)
		self.endRR()
	def addHINFO(self, name, klass, ttl, cpu, os):
		self.addRRheader(name, Type.HINFO, klass, ttl)
		self.addstring(cpu)
		self.addstring(os)
		self.endRR()
	def addMX(self, name, klass, ttl, preference, exchange):
		self.addRRheader(name, Type.MX, klass, ttl)
		self.add16bit(preference)
		self.addname(exchange)
		self.endRR()
	def addNS(self, name, klass, ttl, nsdname):
		self.addRRheader(name, Type.NS, klass, ttl)
		self.addname(nsdname)
		self.endRR()
	def addPTR(self, name, klass, ttl, ptrdname):
		self.addRRheader(name, Type.PTR, klass, ttl)
		self.addname(ptrdname)
		self.endRR()
	def addSOA(self, name, klass, ttl,
			  mname, rname, serial, refresh, retry, expire, minimum):
		self.addRRheader(name, Type.SOA, klass, ttl)
		self.addname(mname)
		self.addname(rname)
		self.add32bit(serial)
		self.add32bit(refresh)
		self.add32bit(retry)
		self.add32bit(expire)
		self.add32bit(minimum)
		self.endRR()
	def addTXT(self, name, klass, ttl, list):
		self.addRRheader(name, Type.TXT, klass, ttl)
		if type(list) in (str, bytes):
			list = [list]
		for txtdata in list:
			self.addstring(txtdata)
		self.endRR()
	# Internet specific RRs (section 3.4) -- class = IN
	def addA(self, name, klass, ttl, address):
		self.addRRheader(name, Type.A, klass, ttl)
		self.addaddr(address)
		self.endRR()
	def addWKS(self, name, ttl, address, protocol, bitmap):
		self.addRRheader(name, Type.WKS, Class.IN, ttl)
		self.addaddr(address)
		self.addbyte(chr(protocol))
		self.addbytes(bitmap)
		self.endRR()
	def addSRV(self):
		raise NotImplementedError

def prettyTime(seconds):
	if seconds<60:
		return seconds,"%d seconds"%(seconds)
	if seconds<3600:
		return seconds,"%d minutes"%(seconds/60)
	if seconds<86400:
		return seconds,"%d hours"%(seconds/3600)
	if seconds<604800:
		return seconds,"%d days"%(seconds/86400)
	else:
		return seconds,"%d weeks"%(seconds/604800)


class RRunpacker(Unpacker):
	def __init__(self, buf):
		Unpacker.__init__(self, buf)
		self.rdend = None
	def getRRheader(self):
		if PY_MAJOR_VERSION >= 3:
			name = self.getname()
		else:
			name = self.getname()
				
		rrtype = self.get16bit()
		klass = self.get16bit()
		ttl = self.get32bit()
		rdlength = self.get16bit()
		self.rdend = self.offset + rdlength
		return (name, rrtype, klass, ttl, rdlength)
	def endRR(self):
		if self.offset != self.rdend:
			raise UnpackError('end of RR not reached')
	def getCNAMEdata(self):
		return self.getname()
	def getHINFOdata(self):
		return self.getstring(), self.getstring()
	def getMXdata(self):
		return self.get16bit(), self.getname()
	def getNSdata(self):
		return self.getname()
	def getPTRdata(self):
		return self.getname()
	def getSOAdata(self):
		return self.getname(), \
			   self.getname(), \
			   ('serial',)+(self.get32bit(),), \
			   ('refresh ',)+prettyTime(self.get32bit()), \
			   ('retry',)+prettyTime(self.get32bit()), \
			   ('expire',)+prettyTime(self.get32bit()), \
			   ('minimum',)+prettyTime(self.get32bit())
	def getTXTdata(self):
		list = []
		while self.offset != self.rdend:
			list.append(self.getstring())
		return list
	def getAdata(self):
		return self.getaddr()
	def getWKSdata(self):
		address = self.getaddr()
		protocol = ord(self.getbyte())
		bitmap = self.getbytes(self.rdend - self.offset)
		return address, protocol, bitmap
	def getSRVdata(self):
		"""
		_Service._Proto.Name TTL Class SRV Priority Weight Port Target
		"""
		priority = self.get16bit()
		weight = self.get16bit()
		port = self.get16bit()
		target = self.getname()
		#print '***priority, weight, port, target', priority, weight, port, target
		return priority, weight, port, target


# Pack/unpack Message Header (section 4.1)

class Hpacker(Packer):
	def addHeader(self, id, qr, opcode, aa, tc, rd, ra, z, rcode,
			  qdcount, ancount, nscount, arcount):
		self.add16bit(id)
		self.add16bit((qr&1)<<15 | (opcode&0xF)<<11 | (aa&1)<<10
				  | (tc&1)<<9 | (rd&1)<<8 | (ra&1)<<7
				  | (z&7)<<4 | (rcode&0xF))
		self.add16bit(qdcount)
		self.add16bit(ancount)
		self.add16bit(nscount)
		self.add16bit(arcount)

class Hunpacker(Unpacker):
	def getHeader(self):
		id = self.get16bit()
		flags = self.get16bit()
		qr, opcode, aa, tc, rd, ra, z, rcode = (
				  (flags>>15)&1,
				  (flags>>11)&0xF,
				  (flags>>10)&1,
				  (flags>>9)&1,
				  (flags>>8)&1,
				  (flags>>7)&1,
				  (flags>>4)&7,
				  (flags>>0)&0xF)
		qdcount = self.get16bit()
		ancount = self.get16bit()
		nscount = self.get16bit()
		arcount = self.get16bit()
		return (id, qr, opcode, aa, tc, rd, ra, z, rcode,
				  qdcount, ancount, nscount, arcount)


# Pack/unpack Question (section 4.1.2)

class Qpacker(Packer):
	def addQuestion(self, qname, qtype, qclass):
		self.addname(qname)
		self.add16bit(qtype)
		self.add16bit(qclass)

class Qunpacker(Unpacker):
	def getQuestion(self):
		return self.getname(), self.get16bit(), self.get16bit()


# Pack/unpack Message(section 4)
# NB the order of the base classes is important for __init__()!

class Mpacker(RRpacker, Qpacker, Hpacker):
	pass

class Munpacker(RRunpacker, Qunpacker, Hunpacker):
	pass


# Routines to print an unpacker to stdout, for debugging.
# These affect the unpacker's current position!

def dumpM(u):
	print('HEADER:')
	(id, qr, opcode, aa, tc, rd, ra, z, rcode,
			  qdcount, ancount, nscount, arcount) = u.getHeader()
	print('id=%d,' % id)
	print('qr=%d, opcode=%d, aa=%d, tc=%d, rd=%d, ra=%d, z=%d, rcode=%d,' \
			  % (qr, opcode, aa, tc, rd, ra, z, rcode))
	if tc: print('*** response truncated! ***')
	if rcode: print('*** nonzero error code! (%d) ***' % rcode)
	print('  qdcount=%d, ancount=%d, nscount=%d, arcount=%d' \
			  % (qdcount, ancount, nscount, arcount))
	for i in range(qdcount):
		print('QUESTION %d:' % i)
		dumpQ(u)
	for i in range(ancount):
		print('ANSWER %d:' % i)
		dumpRR(u)
	for i in range(nscount):
		print('AUTHORITY RECORD %d:' % i)
		dumpRR(u)
	for i in range(arcount):
		print('ADDITIONAL RECORD %d:' % i)
		dumpRR(u)

class DnsResult:

	def __init__(self,u,args):
		self.header={}
		self.questions=[]
		self.answers=[]
		self.authority=[]
		self.additional=[]
		self.args=args
		self.storeM(u)

	def show(self):
		import time
		print('; <<>> PDG.py 1.0 <<>> %s %s'%(self.args['name'],
			self.args['qtype']))
		opt=""
		if self.args['rd']:
			opt=opt+'recurs '
		h=self.header
		print(';; options: '+opt)
		print(';; got answer:')
		print(';; ->>HEADER<<- opcode %s, status %s, id %d'%(
			h['opcode'],h['status'],h['id']))
		flags=list(filter(lambda x,h=h:h[x],('qr','aa','rd','ra','tc')))
		print(';; flags: %s; Ques: %d, Ans: %d, Auth: %d, Addit: %d'%(
			" ".join(flags),h['qdcount'],h['ancount'],h['nscount'],
			h['arcount']))
		print(';; QUESTIONS:')
		for q in self.questions:
			print(';;	  %s, type = %s, class = %s'%(q['qname'],q['qtypestr'],
				q['qclassstr']))
		print()
		print(';; ANSWERS:')
		for a in self.answers:
			print('%-20s	%-6s  %-6s  %s'%(a['name'],repr(a['ttl']),a['typename'],
				a['data']))
		print()
		print(';; AUTHORITY RECORDS:')
		for a in self.authority:
			print('%-20s	%-6s  %-6s  %s'%(a['name'],repr(a['ttl']),a['typename'],
				a['data']))
		print()
		print(';; ADDITIONAL RECORDS:')
		for a in self.additional:
			print('%-20s	%-6s  %-6s  %s'%(a['name'],repr(a['ttl']),a['typename'],
				a['data']))
		print()
		if 'elapsed' in self.args:
			print(';; Total query time: %d msec'%self.args['elapsed'])
		print(';; To SERVER: %s'%(self.args['server']))
		print(';; WHEN: %s'%time.ctime(time.time()))

	def storeM(self,u):
		(self.header['id'], self.header['qr'], self.header['opcode'],
		  self.header['aa'], self.header['tc'], self.header['rd'],
		  self.header['ra'], self.header['z'], self.header['rcode'],
		  self.header['qdcount'], self.header['ancount'],
		  self.header['nscount'], self.header['arcount']) = u.getHeader()
		self.header['opcodestr']=Opcode.opcodestr(self.header['opcode'])
		self.header['status']=Status.statusstr(self.header['rcode'])
		for i in range(self.header['qdcount']):
			#print 'QUESTION %d:' % i,
			self.questions.append(self.storeQ(u))
		for i in range(self.header['ancount']):
			#print 'ANSWER %d:' % i,
			self.answers.append(self.storeRR(u))
		for i in range(self.header['nscount']):
			#print 'AUTHORITY RECORD %d:' % i,
			self.authority.append(self.storeRR(u))
		for i in range(self.header['arcount']):
			#print 'ADDITIONAL RECORD %d:' % i,
			self.additional.append(self.storeRR(u))

	def storeQ(self,u):
		q={}
		q['qname'], q['qtype'], q['qclass'] = u.getQuestion()
		q['qtypestr']=Type.typestr(q['qtype'])
		q['qclassstr']=Class.classstr(q['qclass'])
		return q

	def storeRR(self,u):
		r={}
		r['name'],r['type'],r['class'],r['ttl'],r['rdlength'] = u.getRRheader()
		r['typename'] = Type.typestr(r['type'])
		r['classstr'] = Class.classstr(r['class'])
		#print 'name=%s, type=%d(%s), class=%d(%s), ttl=%d' \
		#	  % (name,
		#		type, typename,
		#		klass, Class.classstr(class),
		#		ttl)
		mname = 'get%sdata' % r['typename']
		if hasattr(u, mname):
			r['data']=getattr(u, mname)()
		else:
			r['data']=u.getbytes(r['rdlength'])
		return r

def dumpQ(u):
	qname, qtype, qclass = u.getQuestion()
	print('qname=%s, qtype=%d(%s), qclass=%d(%s)' \
			  % (qname,
				 qtype, Type.typestr(qtype),
				 qclass, Class.classstr(qclass)))

def dumpRR(u):
	name, type, klass, ttl, rdlength = u.getRRheader()
	typename = Type.typestr(type)
	print('name=%s, type=%d(%s), class=%d(%s), ttl=%d' \
			  % (name,
				 type, typename,
				 klass, Class.classstr(klass),
				 ttl))
	mname = 'get%sdata' % typename
	if hasattr(u, mname):
		print('  formatted rdata:', getattr(u, mname)())
	else:
		print('  binary rdata:', u.getbytes(rdlength))

if __name__ == "__main__":
	testpacker()
#
# $Log: Lib.py,v $
# Revision 1.11  2002/03/19 13:05:02  anthonybaxter
# converted to class based exceptions (there goes the python1.4 compatibility :)
#
# removed a quite gross use of 'eval()'.
#
# Revision 1.10  2002/03/19 12:41:33  anthonybaxter
# tabnannied and reindented everything. 4 space indent, no tabs.
# yay.
#
# Revision 1.9  2002/03/19 10:30:33  anthonybaxter
# first round of major bits and pieces. The major stuff here (summarised
# from my local, off-net CVS server :/ this will cause some oddities with
# the
#
# tests/testPackers.py:
#   a large slab of unit tests for the packer and unpacker code in DNS.Lib
#
# DNS/Lib.py:
#   placeholder for addSRV.
#   added 'klass' to addA, make it the same as the other A* records.
#   made addTXT check for being passed a string, turn it into a length 1 list.
#   explicitly check for adding a string of length > 255 (prohibited).
#   a bunch of cleanups from a first pass with pychecker
#   new code for pack/unpack. the bitwise stuff uses struct, for a smallish
#	 (disappointly small, actually) improvement, while addr2bin is much
#	 much faster now.
#
# DNS/Base.py:
#   added DiscoverNameServers. This automatically does the right thing
#	 on unix/ win32. No idea how MacOS handles this.  *sigh*
#	 Incompatible change: Don't use ParseResolvConf on non-unix, use this
#	 function, instead!
#   a bunch of cleanups from a first pass with pychecker
#
# Revision 1.8  2001/08/09 09:08:55  anthonybaxter
# added identifying header to top of each file
#
# Revision 1.7  2001/07/19 07:50:44  anthony
# Added SRV (RFC 2782) support. Code from Michael Strer.
#
# Revision 1.6  2001/07/19 07:39:18  anthony
# 'type' -> 'rrtype' in getRRheader(). Fix from Michael Strer.
#
# Revision 1.5  2001/07/19 07:34:19  anthony
# oops. glitch in storeRR (fixed now).
# Reported by Bastian Kleineidam and by greg lin.
#
# Revision 1.4  2001/07/19 07:16:42  anthony
# Changed (opcode&0xF)<<11 to (opcode*0xF)<<11.
# Patch from Timothy J. Miller.
#
# Revision 1.3  2001/07/19 06:57:07  anthony
# cvs keywords added
#
#



from . import killtree
import time
from subprocess import Popen, PIPE
import threading

class Puppet:
	counter = 0
	def __init__ (self, logger = None, communicate = True):
		self.logger = logger
		self.p = None
		self.__lock = threading.Lock ()
		self.__active = False
		self.__last_activated = time.time ()
		self.__communicate = communicate		
		self.counter += 1
	
	def __str__ (self):	
		return 'Puppet #%d' % self.counter
		
	def set_active (self, flag):
		with self.__lock:
			self.__active = flag
			if flag == False:
				self.p = None
		
	def is_active (self):
		with self.__lock:
			r = self.__active		
		return r
			
	def start (self, command):
		self.set_active (True)
		threading.Thread (target = self.threaded_run, args = (command,)).start ()
		
	def threaded_run (self, command):
		try:
			self.create_process (command)
		finally:
			self.wait ()

	def kill (self):
		if not (self.is_active and self.p):
			self.set_active (False)
			return
		killtree.kill (self.p.pid)
		
	def wait (self):
		with self.__lock:
			p = self.p
			
		if not p:
			self.log ("[error] -- terminated with -1")
				
		else:
			while 1:
				try:
					with self.__lock:					
						exitcode = self.p.poll ()
				except AttributeError:	
					exitcode = -1
					break										
				if exitcode is not None:
					break
				time.sleep (1)
			self.log ("[info] -- terminated with %s" % exitcode)		
		
		self.set_active (False)
		
	def is_timeout (self, timeout):
		return time.time () - self.__last_activated > timeout
	
	def set_last_activate (self):
		self.__last_activated = time.time ()
			
	def log (self, line):
		if line[0].isdigit ():
			# auqests.lib.logger classes
			line = line [20:].strip ()
		if self.logger:
			self.logger (line, "")			
		self.set_last_activate ()
			
	def create_process (self, cmd):				
		self.log ("[info] -- start process: %s" % " ".join (cmd))
		s_time = time.time ()
		self.__lock.acquire ()
		try:
			self.p = Popen (
				cmd,
		    universal_newlines=True,
		    stdout=self.__communicate and PIPE or None, stderr=self.__communicate and PIPE or None,
		    shell = False
			)
		finally:
			self.__lock.release ()
		
		if not self.__communicate:
			return
			
		for line in iter (self.p.stdout.readline, ''):
			self.log (line)

		self.p.stdout.close ()			
		e = self.p.stderr.read ()
		if e: self.log (e)		
		self.p.stderr.close ()
	
	


import boto.ec2
import time
from boto.ec2.elb import ELBConnection
from . import iquery, metrics

class LoadBalancer:
	def __init__ (self, name, region = "us-east-1"):
		self.name = name
		self.region = region
		self.conn = boto.ec2.connect_to_region (region)
		self.image_id = None
		
	def create_instance (self, param):
		self.image_id = param ['image_id']		
		param ['monitoring_enabled'] = False
		param ['monitoring_enabled'] = False
		param ['dry_run'] = False
	
		reservation = self.conn.run_instances (**param)
		instance = reservation.instances [0]
		while instance.update() == 'pending':
			print ('waiting...')
			time.sleep(10)	
		
		status = instance.update()
		if status == 'running':
			instance.add_tag("Name", self.name + "-" + self.image_id [4:])
			instance.add_tag("ami-id", self.image_id)
			instance.add_tag("system", "skitai")
		else:
			print('instance status: ' + status)		
		
		return instance
			
	def register_to_elb (self, instance_id):
		elb = ELBConnection()
		elb.register_instances (self.name, instance_id)
	
	def deregister_to_elb (self, instance_id):
		elb = ELBConnection()
		elb.deregister_instances (self.name, instance_id)
			
	def terminate (self, instance_id):
		instance = iquery.get_instance_by_id (instance_id)
		instance.terminate ()
	
	def get_instance_id (self):
		instance = [i for i in iquery.get_instances_by_tag ("system", 'skitai') if i.update () == "running"][0]
		instance = iquery.get_instance_by_id (instance.id)
		return instance.id
	
	def get_status (self):
		g = {}
		for h in self.get_health ():
			g [h.instance_id] = (h.state, metrics.get_latest (h.instance_id))
		return g
	
	def get_health (self):
		elb = ELBConnection()
		this = elb.get_all_load_balancers ([self.name])[0]
		return this.get_instance_health ()			
		
	def get_instance_ids (self):
		return [h.instance_id for h in self.get_health ()]



from .protocols.http import request as http_request, request_handler as http_request_handler
from .protocols.ws import request_handler as ws_request_handler, request as ws_request
from .protocols.grpc import request as grpc_request
from .dbapi import request as dbo_request
from .protocols.proxy import tunnel_handler
from .protocols.http import localstorage as ls

def make_ws (_method, url, params, auth, headers, meta, proxy, logger):
	req = ws_request.Request (url, params, headers, auth, logger, meta)
	if proxy:
		if _method == 'wss':
			handler_class = tunnel_handler.WSSSLProxyTunnelHandler
		else:
			handler_class = tunnel_handler.WSProxyTunnelHandler
	else:		
		handler_class = ws_request_handler.RequestHandler
	req.handler = handler_class	
	return req

CONTENT_TYPES = [
	('xml', "text/xml", "text/xml"),
	('json', "application/json", "application/json"),
]
	
def make_http (_method, url, params, auth, headers, meta, proxy, logger):
	global CONTENT_TYPES
	if not headers:
		headers = {}
	if ls.g:
		headers ['Cookie'] = ls.g.get_cookie_as_string (url)
	if proxy and url.startswith ('https://'):
		handler_class = tunnel_handler.SSLProxyTunnelHandler
	else:			
		handler_class = http_request_handler.RequestHandler

	if _method == "rpc":
		rpcmethod, params = params
		req = http_request.XMLRPCRequest (url, rpcmethod, params, headers, auth, logger, meta)
		
	elif _method == "grpc":
		rpcmethod, params = params
		req = grpc_request.GRPCRequest (url, rpcmethod, params, headers, auth, logger, meta)
	
	elif _method == "upload":
		req = http_request.HTTPMultipartRequest (url, "POST", params, headers, auth, logger, meta)
				
	else:
		ct = "application/x-www-form-urlencoded"
		ac = "*/*"
		for _alias, _ct, _ac in CONTENT_TYPES:
			if _method.endswith (_alias):
				_method = _method [:-len (_alias)]				
				ct = _ct
				ac = _ac
				break
		
		headers ['Accept'] = ac
		if params:						
			headers ['Content-Type'] = ct
		req = http_request.HTTPRequest (url, _method.upper (), params, headers, auth, logger, meta)		
	
	req.handler = handler_class
	return req

def make_dbo (_method, server, dbmethod, params, dbname, auth, meta, logger):
	return dbo_request.Request (_method [1:], server, dbname, auth, dbmethod, params, None, meta)



import sys, asyncore, time
import gc
import select
import os
import bisect
import socket
import time
import types
try:
	from pympler import muppy, summary, tracker
except ImportError:
	pass	

if os.name == "nt":
	from errno import WSAENOTSOCK
	
_shutdown_phase = 0
_shutdown_timeout = 30 # seconds per phase
_exit_code = 0
_last_maintern = 0
_maintern_interval = 3.0
_killed_zombies = 0
_select_errors = 0
_poll_count = 0
_polling = 0
_logger = None
	
class Maintern:
	def __init__ (self):
		self.q = []
		
	def sched (self, interval, func, args = None):
		now = time.time ()
		self.q.append ((now + interval, interval, func, args))
		self.q.sort (key = lambda x: x [0])
		#bisect.insort (self.q, (now + interval, interval, func, args))
	
	def __call__ (self, now):
		excutes = 0
		for exetime, interval, func, args in self.q:
			if exetime > now: break
			excutes += 1
			if args: 
				func (now, *args)
			else:
				func (now)

		for i in range (excutes):
			exetime, interval, func, args = self.q.pop (0)
			#bisect.insort (self.q, (now + interval, interval, func, args))
			self.q.append ((now + interval, interval, func, args))
			self.q.sort (key = lambda x: x [0])
		
def maintern_gc (now):
	gc.collect ()
	
def maintern_zombie_channel (now):
	global _killed_zombies
		
	for channel in list(asyncore.socket_map.values()):
		if hasattr (channel, "handle_timeout"):
			try:
				# +3 is make gap between server & client
				iszombie = (now - channel.event_time) > channel.zombie_timeout
			except AttributeError:
				continue
			if iszombie:				
				_killed_zombies += 1
				try:
					channel.handle_timeout ()
				except:
					channel.handle_error ()

maintern = None
def init (kill_zombie_interval = 10.0, logger = None):
	global maintern, _logger
	
	_logger = logger
	maintern = Maintern ()
	maintern.sched (kill_zombie_interval, maintern_zombie_channel)
	maintern.sched (300.0, maintern_gc)

summary_tracker = None
def summary_track (now):
	global summary_tracker
	
	all_objects = muppy.get_objects ()
	#all_objects = muppy.filter (all_objects, Type = dict)
	#for each in all_objects:
	#	print (each)
	sum1 = summary.summarize (all_objects)
	summary.print_ (sum1)
	print ('-' * 79)	
	if summary_tracker is None:
		summary_tracker = tracker.SummaryTracker ()
	summary_tracker.print_diff ()	
	
def enable_memory_track (interval = 10.0):
	global maintern	
	maintern.sched (interval, summary_track)	

def shutdown (exit_code, shutdown_timeout = 30.0):
	global _shutdown_phase
	global _shutdown_timeout
	global _exit_code
	
	if _shutdown_phase:
		# aready entered
		return
		
	if _shutdown_phase == 0:
		_exit_code = exit_code
		_shutdown_phase = 1
		
	_shutdown_timeout = shutdown_timeout

def loop (timeout = 30.0):
	global _shutdown_phase
	global _shutdown_timeout
	global _exit_code
	global _polling
	global maintern
	
	if maintern is None:
		init ()
		
	_shutdown_phase = 0
	_shutdown_timeout = 30
	_exit_code = 0	
	_polling = 1
	
	try: 
		lifetime_loop(timeout)
	except KeyboardInterrupt:
		graceful_shutdown_loop()
	else:
		graceful_shutdown_loop()
	
	_polling = 0
	
if hasattr(select, 'poll'):
	poll_fun = asyncore.poll2
else:
	poll_fun = asyncore.poll

def remove_notsocks (map):
	global _select_errors, _logger
	
	# on Windows we can get WSAENOTSOCK if the client
	# rapidly connect and disconnects
	killed = 0
	for fd in list(map.keys()):
		try:
			select.select([fd], [], [], 0)
			
		except:
			_logger and _logger.trace ()
			killed += 1
			_select_errors += 1
						
			try:
				obj = map [fd]
			except KeyError:
				pass
			else:
				try:
					try: obj.handle_expt ()
					except: obj.handle_error ()
				except:
					_logger and _logger.trace ()
					
				try: del map [fd]
				except KeyError: pass
					
	return killed

def poll_fun_wrap (timeout, map = None):
	global _logger
	
	if map is None:
		map = asyncore.socket_map
		
	try:		
		poll_fun (timeout, map)
	
	except (TypeError, OSError) as why:
		# WSAENOTSOCK
		remove_notsocks (map)
	
	except ValueError:
		# negative file descriptor, testing all sockets
		killed = remove_notsocks (map)
		# or too many file descriptors in select(), divide and conquer
		if not killed:			
			half = int (len (map) / 2)
			tmap = {}
			cc = 0			
			for k, v in list(map.items ()):
				tmap [k] = v
				cc += 1
				if cc == half:
					poll_fun_wrap (timeout, tmap)
					tmap = {}
			poll_fun_wrap (timeout, tmap)
	
	except:
		_logger and _logger.trace ()
		raise
	
def lifetime_loop (timeout = 30.0, count = 0):
	global _last_maintern
	global _maintern_interval

	map = asyncore.socket_map
	loop = 0
	while map and _shutdown_phase == 0:		
		poll_fun_wrap (timeout, map)
		now = time.time()
		if (now - _last_maintern) > _maintern_interval:
			maintern (now)
			_last_maintern = time.time ()
		loop += 1
		if count and loop > count:
			break	
		
def graceful_shutdown_loop ():
	global _shutdown_phase
	timestamp = time.time()
	timeout = 1.0
	map = asyncore.socket_map	
	while map and _shutdown_phase < 4:
		time_in_this_phase = time.time() - timestamp
		veto = 0
		for fd,obj in list(map.items()):
			try:
				fn = getattr (obj,'clean_shutdown_control')
			except AttributeError:
				pass
			else:
				try:
					veto = veto or fn (_shutdown_phase, time_in_this_phase)
				except:					
					obj.handle_error()
					
		if veto and time_in_this_phase < _shutdown_timeout:
			poll_fun_wrap (timeout, map)					
		else:
			_shutdown_phase += 1
			timestamp = time.time()



import os

if os.name == "nt":
	import win32process
	import win32event
	import pywintypes
	
	def execute (cmd, timeout = 0):
		if timeout == 0:
			timeout = win32event.INFINITE
			
		info  = win32process.CreateProcess(None, cmd, None, None, 0, 0, None, None, win32process.STARTUPINFO())
		subprocess = info [0]
		
		rc = win32event.WaitForSingleObject (subprocess, timeout)			
		
		if rc == win32event.WAIT_FAILED:	
			return -1
			
		if rc == win32event.WAIT_TIMEOUT:
			try:
				win32process.TerminateProcess (subprocess, 0)					
			except pywintypes.error:
				return -3
			return -2
		
		if rc == win32event.WAIT_OBJECT_0:
			return win32process.GetExitCodeProcess(subprocess)
	
